
/**
 * Implements the quicksort algorithm.
 *  @author Koffman and Wolfgang
 **/
public class QuickSort implements SortAlgorithm {

    /**
     * Sort the table using the quicksort algorithm.
     * @pre table contains Comparable objects.
     * @post table is sorted.
     * @param table The array to be sorted
     */
    @Override
    public <T extends Comparable<T>>  void sort(T[] table) {
        // Sort the whole table.
        quickSort(table, 0, table.length - 1);
    }

    /**
     * Sort a part of the table using the quicksort algorithm.
     * @post The part of table from first through last is sorted.
     * @param table The array to be sorted
     * @param first The index of the low bound
     * @param last The index of the high bound
     */
    protected <T extends Comparable<T>>  void quickSort(T[] table, int first, int last) {
        if (first < last) { // There is data to be sorted.
            // Partition the table.
            int pivIndex = partition(table, first, last);
            // Sort the left half.
            quickSort(table, first, pivIndex - 1);
            // Sort the right half.
            quickSort(table, pivIndex + 1, last);
        }
    }

    /**
     * Partition the table so that values from first to pivIndex
     * are less than or equal to the pivot value, and values from
     * pivIndex to last are greater than the pivot value.
     * @param table The table to be partitioned
     * @param first The index of the low bound
     * @param last  The index of the high bound
     * @return The location of the pivot value
     */
    protected <T extends Comparable<T>>  int partition(T[] table, int first, int last) {
        // Put the median of table[first], table[middle], table[last]
        // int table[first] and use this value as the pivot
        int middle = first + (last - first) / 2;
        bubbleSort3(table, first, middle, last);
        swap(table, first, middle);
        T pivot = table[first];
        int up = first;
        int down = last;
        do {
            // Invariant:
            // All items in table[first . . . up - 1] <= pivot
            // All items in table[down + 1 . . . last] > pivot
            while ((up < last) && (pivot.compareTo(table[up]) >= 0)) {
                up++;
            }
            // assert: up equals last or table[up] > pivot.
            while (pivot.compareTo(table[down]) < 0) {
                down--;
            }
            // assert: down equals first or table[down] <= pivot.
            if (up < down) { // if up is to the left of down.
                // Exchange table[up] and table[down].
                swap(table, up, down);
            }
        } while (up < down); // Repeat while up is left of down.

        // Exchange table[first] and table[down] thus putting the
        // pivot value where it belongs.
        swap(table, first, down);

        // Return the index of the pivot value.
        return down;
    }

    /**
     * Sort table[first], table[middle], table[last].
     * @param table The table to be sorted
     * @param first Index of the first element
     * @param last Index of the last element
     */
    protected <T extends Comparable<T>>  void bubbleSort3(T[] table, int first, 
                                                          int middle, int last) {
        if ((table[middle].compareTo(table[first]) < 0)) {
            swap(table, first, middle);
        }
        // assert: table[first] <= table[middle]
        if ((table[last].compareTo(table[middle])) < 0) {
            swap(table, middle, last);
        }
        // assert: table[last] is the largest of the three.
        if ((table[middle].compareTo(table[first]) < 0)) {
            swap(table, first, middle);
        }
        // assert table[first] <= table[middle] <= table[last]
    }

    /**
     * Swap the items in table[i] and table[j].
     * @param table The array that contains the items
     * @param i The index of one item
     * @param j The index of the other item
     */
    protected <T extends Comparable<T>>  void swap(T[] table, int i, int j) {
        T temp = table[i];
        table[i] = table[j];
        table[j] = temp;
    }
}
